/**
   This is a simple abstract generic hash table with chaining,
   load factor, and rehashing. This abstract class leaves the
   choice of the hash function up to each implementation.
*/
public abstract class ChainHashTable<K,V> extends HashTable<K,V>
{
   private Node<K,V>[] buckets; // The array of buckets (the hash table).
   public boolean debug = true;

   private class Node<S,T>
   {
      public S key;
      public T value;
      public Node<S,T> next;
      @Override public String toString()
      {
         String result = "<"+key+", "+value+">";
         if (null != next) result += " ==> " + next;
         return result;
      }
   }


   /**
      Constructs a hash table array for this HashTable.

      @param length  the initial length of the hash table array
   */
   public ChainHashTable(int length)
   {
      this.length = length;
      @SuppressWarnings({"unchecked", "rawtypes"})
      final Node<K,V>[] buckets = (Node<K,V>[])new Node[length]; // Need to use a raw type!
      this.buckets = buckets;
      size = 0;
      maxLoadFactor = 0.75; // Default maximum load factor.
   }


   /**
      @param key    key that is hashed into a hash table index
      @param value  value that is placed in the hash table at the hashed index
   */
   public void insert(K key, V value)
   {
      if (loadFactor() >= maxLoadFactor)
      {
         rehash();
      }

      final int hashCode = hashFunction(key);
      int index = hashCode % length; // Make sure the hash-code fits into the array.
      index = Math.abs(index);       // Some hash functions return negative values.

      ++size;
      final Node<K,V> current = buckets[index]; // Find the appropriate bucket.
      final Node<K,V> newNode = new Node<>();   // Add a new node at the front of the linked-list.
      newNode.key  = key;
      newNode.value = value;
      newNode.next = current;
      buckets[index] = newNode;
   }


   /**
      Return a reference to the Value with the given Key.
      Return null if the Key is not found in the hash table.

      @param key  key that is hashed into a hash table index
      @return the value found at the hash table's hashed index
   */
   public V search(K key)
   {
      final int hashCode = hashFunction(key);
      int index = hashCode % length;
      index = Math.abs(index);

      Node<K,V> current = buckets[index]; // Find the appropriate bucket.
      while (current != null) // Search the linked-list.
      {
         if (current.key.equals(key))
         {
            return current.value;
         }
         current = current.next;
      }
      return null; // return null if not found
   }


   /**
      Remove the value with the given key.
      Return false if the key is not found in the hash table.

      @param key  key that is hashed into a hash table index
      @return true if the key was found and the (key,value) pair was deleted
   */
   public boolean delete(K key)
   {
      final int hashCode = hashFunction(key);
      int index = hashCode % length;
      index = Math.abs(hashCode);

      Node<K,V> current = buckets[index]; // Find the appropriate bucket.
      // If the (key,value) pair is in the bucket's linked
      // list, then remove the pair from the list.
      Node<K,V> previous = null;
      while (current != null) // Search the linked-list.
      {
         if (current.key.equals(key)) // Found the key.
         {
            if (previous == null) buckets[index] = current.next;
            else                   previous.next = current.next;
            --size;
            return true;
         }
         previous = current;
         current = current.next;
      }
      return false; // Return false if the key is not found.
   }


   /**
      Double the size the bucket array and then rehash
      every (key, value) pair from the old bucket array
      into the new bucket array.
   */
   private void rehash()
   {
      if (debug){ // Log to stdout each rehashing step.
         System.out.println("Rehash: before");
         System.out.println( this );
      }

      // Save the old hash table.
      final Node<K,V>[] oldBuckets = buckets;
      // Create a new bucket array with twice the length.
      this.length = 2 * this.length;
      @SuppressWarnings({"unchecked", "rawtypes"})
      final Node<K,V>[] buckets = (Node<K,V>[])new Node[this.length]; // Need to use a raw type!
      this.buckets = buckets;
      this.size = 0; // The new bucket array is currently empty.

      // Rehash keys from the old hash table into the new hash table.
      for (int i = 0; i < oldBuckets.length; ++i)
      {
         if (null != oldBuckets[i])
         {
            // Rehash every (key, value) pair in this linked list.
            Node<K,V> current = oldBuckets[i];
            while (current != null)
            {
               insert(current.key, current.value);
               current = current.next;
            }
         }
      }

      if (debug){ // Log to stdout each rehashing step.
         System.out.println("Rehash: after");
         System.out.println( this );
      }
   }


   @Override public String toString()
   {
      String result = "";
      for (int i = 0; i < this.length; ++i)
      {
         result += "["+i+"] " + buckets[i] + "\n";
      }
      return result;
   }
}
