/**
   This is a simple abstract generic hash table with open
   addressing, load factor, and rehashing. This abstract
   class leaves the choice of the hash function up to each
   implementation.
*/
public abstract class OpenHashTable<K,V> extends HashTable<K,V>
{
   private K[] keyBuckets;
   private V[] valueBuckets;
   public boolean debug = true;

   /**
      Constructs a hash table array for this HashTable.

      @param length  the initial length of the hash table array
   */
   @SuppressWarnings("unchecked")
   public OpenHashTable(int length)
   {
      this.length = length;
      keyBuckets   = (K[])new Object[length]; // unchecked cast
      valueBuckets = (V[])new Object[length]; // unchecked cast
      size = 0;
      maxLoadFactor = 0.75; // Default maximum load factor.
   }


   /**
      @param key    key that is hashed into a hash table index
      @param value  value that is placed in the hash table at the hashed index
   */
   public void insert(K key, V value)
   {
      if (loadFactor() >= maxLoadFactor)
      {
         rehash();
      }

      final int hashCode = hashFunction(key);
      int index = hashCode % length; // Make sure the hash-code fits into the array.
      index = Math.abs(index);       // Some hash functions return negative values.

      // Find the first available empty bucket.
      // Since the load factor must be < 1 there
      // must be an empty slot.
      while (null != keyBuckets[index])
      {
         index = (index + 1) % length; // sequential probing
      }
      // Put the (key, value) pair in the empty bucket.
        keyBuckets[index] = key;
      valueBuckets[index] = value;
      ++size;
   }


   /**
      Return a reference to the Value with the given Key.
      Return null if the Key is not found in the hash table.

      @param key  key that is hashed into a hash table index
      @return the value found at the hash table's hashed index
   */
   public V search(K key)
   {
      final int hashCode = hashFunction(key);
      int index = hashCode % length;
      index = Math.abs(index);

      // Sequentially probe for the given key.
      int probes = 0;
      while ( probes < this.length                 // stop after length probes
           && ( keyBuckets[index] == null          // or
             || ! keyBuckets[index].equals(key)) ) // stop when the key is found
      {
         index = (index + 1) % this.length; // sequential probing
         ++probes;
      }
      if (probes < this.length)
      {
         return valueBuckets[index];
      }
      else
      {
         return null; // return null if the key is not found
      }
   }


   /**
      Remove the value with the given key.
      Return false if the key is not found in the hash table.

      @param key  key that is hashed into a hash table index
      @return true if the key was found and the (key,value) pair was deleted
   */
   public boolean delete(K key)
   {
      final int hashCode = hashFunction(key);
      int index = hashCode % length;
      index = Math.abs(hashCode);

      // Sequentially probe for the given key.
      int probes = 0;
      while ( probes < this.length                 // stop after length probes
           && ( keyBuckets[index] == null          // or
             || ! keyBuckets[index].equals(key)) ) // stop when the key is found
      {
         index = (index + 1) % this.length; // sequential probing
         ++probes;
      }
      if (probes < this.length)
      {
           keyBuckets[index] = null;
         valueBuckets[index] = null;
         --size;
         return true;
      }
      else
      {
         return false; // return false if the key is not found
      }
   }


   /**
      Double the size the bucket array and then rehash
      every (key, value) pair from the old bucket array
      into the new bucket array.
   */
   @SuppressWarnings("unchecked")
   private void rehash()
   {
      if (debug){ // Log to stdout each rehashing step.
         System.out.println("Rehash: before");
         System.out.println( this );
      }

      // Save the old hash table.
      final K[]   oldKeys = keyBuckets;
      final V[] oldValues = valueBuckets;
      // Create two new bucket arrays with twice the length.
      this.length = 2 * this.length;
      keyBuckets   = (K[])new Object[this.length]; // unchecked cast
      valueBuckets = (V[])new Object[this.length]; // unchecked cast
      this.size = 0; // The new bucket arrays are currently empty.

      // Rehash keys from the old hash table into the new hash table.
      for (int i = 0; i < oldKeys.length; ++i)
      {
         if (null != oldKeys[i])
         {
            // Rehash this (key, value) pair.
            insert(oldKeys[i], oldValues[i]);
         }
      }

      if (debug){ // Log to stdout each rehashing step.
         System.out.println("Rehash: after");
         System.out.println( this );
      }
   }


   @Override public String toString()
   {
      String result = "";
      for (int i = 0; i < this.length; ++i)
      {
         result += "["+i+"] " + "<" + keyBuckets[i] + ", " + valueBuckets[i] + ">" + "\n";
      }
      return result;
   }
}
