import java.util.AbstractSet;
import java.util.Iterator;
import java.util.NoSuchElementException;

/**
   A HashSet<E> stores an unordered collection of E objects,
   using a hash table.
*/
//@SuppressWarnings("unchecked")
public class HashSet<E> extends AbstractSet<E>
{
   private Object[] buckets; // The hash table.
   private int size;         // The number of items in the hash table.

   private class Node
   {
      public E data;
      public Node next;
   }


   /**
      Constructs a hash table for the HashSet<E>.
      @param bucketsLength the length of the buckets array
   */
   public HashSet(int bucketsLength)
   {
      buckets = new Object[bucketsLength];
      size = 0;
   }

   /**
      Tests for set membership.
      @param x an object
      @return true if x is an element of this set
   */
   //@SuppressWarnings("unchecked")
   public boolean contains(Object x)
   {
      int h = x.hashCode();
      if (h < 0) h = -h;
      h = h % buckets.length;

      Node current = (Node)buckets[h];
      while (current != null)
      {
         if (current.data.equals(x)) return true;
         current = current.next;
      }
      return false;
   }

   /**
      Adds an element to this set.
      @param x an object
      @return true if x is a new object, false if x was
      already in the set
   */
   //@SuppressWarnings("unchecked")
   public boolean add(E x)
   {
      int h = x.hashCode();
      if (h < 0) h = -h;
      h = h % buckets.length;

      Node current = (Node)buckets[h];
      while (current != null)
      {
         if (current.data.equals(x))
            return false; // Already in the set
         current = current.next;
      }
      final Node newNode = new Node();
      newNode.data = x;
      newNode.next = (Node)buckets[h];
      buckets[h] = newNode;
      size++;
      return true;
   }

   /**
      Removes an object from this set.
      @param x an object
      @return true if x was removed from this set, false
      if x was not an element of this set
   */
   //@SuppressWarnings("unchecked")
   public boolean remove(Object x)
   {
      int h = x.hashCode();
      if (h < 0) h = -h;
      h = h % buckets.length;

      Node current = (Node)buckets[h];
      Node previous = null;
      while (current != null)
      {
         if (current.data.equals(x))
         {
            if (previous == null) buckets[h] = current.next;
            else previous.next = current.next;
            size--;
            return true;
         }
         previous = current;
         current = current.next;
      }
      return false;
   }

   /**
      Gets the number of elements in this set.
      @return the number of elements
   */
   public int size()
   {
      return size;
   }

   /**
      Returns an iterator that traverses the elements of this set.
      @return a hash set iterator
   */
   public Iterator<E> iterator()
   {
      return new HashSetIterator();
   }


   /**
      Inner class.
   */
   private class HashSetIterator implements Iterator<E>
   {
      private int bucket;
      private Node current;
      private int previousBucket;
      private Node previous;

      /**
         Constructs a hash set iterator that points to the
         first element of the hash set.
      */
      public HashSetIterator()
      {
         current = null;
         bucket = -1;
         previous = null;
         previousBucket = -1;
      }

      public boolean hasNext()
      {
         if (current != null && current.next != null)
            return true;
         for (int b = bucket + 1; b < buckets.length; b++)
            if (buckets[b] != null) return true;
         return false;
      }

      //@SuppressWarnings("unchecked")
      public E next()
      {
         previous = current;
         previousBucket = bucket;
         if (current == null || current.next == null)
         {
            // Move to next bucket
            bucket++;

            while (bucket < buckets.length
                  && buckets[bucket] == null)
               bucket++;
            if (bucket < buckets.length)
               current = (Node)buckets[bucket];
            else
               throw new NoSuchElementException();
         }
         else // Move to next element in bucket
            current = current.next;
         return current.data;
      }

      public void remove()
      {
         if (previous != null && previous.next == current)
            previous.next = current.next;
         else if (previousBucket < bucket)
            buckets[bucket] = current.next;
         else
            throw new IllegalStateException();
         current = previous;
         bucket = previousBucket;
      }
   }
}
