/**
   Here is a brief outline of the pseudo code for a generic hash table.

   This pseudo code does not handle collisions. That makes this
   implementation impractical. But in the extreme case of a very
   large hash table array and a very good hash function, this
   implementation would work. In that case, every operation
   would have complexity O(1).

   This pseudo code delegates the "hash function" to each
   object's hashCode() method from Java's Object class.
*/
abstract class hashtable<K,V>
{
   private int length;  // Length of the bucket array.
   @SuppressWarnings("unchecked")
   private V[] hashArray = (V[])new Object[length]; // The array of buckets.
   private int size;  // Number of values stored in this hash table.

   private int hashFunction(K key)
   {
      return key.hashCode(); // Let the object compute its own hash value.
   }

   /**
      @param key    key that is hashed into a hash table index
      @param value  value that is placed in the hash table at the hashed index
   */
   public void insert(K key, V value)
   {
      final int hashCode = hashFunction(key);
      int index = hashCode % length; // Make sure the hash-code fits into the array.
      index = Math.abs(index);       // Some hash functions return negative values.

      hashArray[index] = value; // Ignore collisions.
      ++size;
   }


   /**
      Return a reference to the Value with the given Key.
      Return null if the Key is not found in the hash table.

      @param key  key that is hashed into a hash table index
      @return the value found at the hash table's hashed index
   */
   public V search(K key)
   {
      final int hashCode = hashFunction(key);
      int index = hashCode % length;
      index = Math.abs(index);

      return hashArray[index]; // returns null if the key is not being used
   }


   /**
      Remove the value with the given key.
      Return false if the key is not found
      in the hash table.

      @param key  key that is hashed into a hash table index
      @return true if the key was found and the (key,value) pair was deleted
   */
   public boolean delete(K key)
   {
      final int hashCode = hashFunction(key);
      int index = hashCode % length;
      index = Math.abs(hashCode);

      if ( hashArray[index] != null ) // We are ignoring collisions.
      {
         hashArray[index] = null;
         --size;
         return true;
      }
      else
      {
         return false;
      }
   }


   /**
      @return the number of (key, value) pairs stored in this hash table
   */
   public int size()
   {
      return size;
   }
}
